# Copyright (c) 2023 Amphion.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .new_trainer import BaseTrainer
from .new_inference import BaseInference
